//---------------------------------------------------------------------------------
// File Name: DemoCars.cpp
// Description: 	
//
// Copyright (C) 2004 - 2006 True Axis Pty Ltd, Australia. 
// All Rights Reserved.
//
// Use of this code is covered by the "TRUE AXIS PHYSICS SDK DEMOS SOURCE LICENSE AGREEMENT".
// Details of this agreement are found in the file "TrueAxisPhysicsSDKDemosSourceLicence.txt".
// You have permission to modify this file and use it for any purpose under the terms of
// the agreement.
//
// History:
//		Created File.
//---------------------------------------------------------------------------------

#include "../Interface/IApp.h"
#include "../Interface/IDemoPlugin.h"
#include "../Interface/IMouse.h"
#include "../Interface/ICamera.h"

#include "../TA/Common/RandFunc.h"
#include "../TA/Common/Geometry.h"

#include "../TA/Physics/Physics.h"
#include "../TA/Physics/CollisionObjectAABBMesh.h"
#include "../TA/Physics/CollisionObjectCombo.h"
#include "../TA/Physics/CollisionObjectConvex.h"
#include "../TA/Physics/DynamicObjectCar.h"
#include "../TA/Physics/CarTemplate.h"

#include "../Shared/Landscape.h"
#include "../Shared/DemoPhysicsRender.h"

#include "RaceTrack.h"
#include "RacingAi.h"

#include <windows.h>

using namespace TA;
	
static bool KeyIsDown(int nKey) { return (GetAsyncKeyState(nKey) & 0x8000) != 0; }
static bool KeyIsJustPressed(int nKey) { return (GetAsyncKeyState(nKey) & 0x8001) == 0x8001; }

class DemoCars : public IDemoPlugin
{
public:
	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	DemoCars()
	{
		m_nNumDynamicObjects = 0;
		m_ppDynamicObjectList = 0;
		m_pRacingAIList = 0;
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	~DemoCars()
	{
		Finalise();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Release() { delete this; }

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Initialise(IApp* pApp)
	{
		m_pApp = pApp;
		m_bHelp = true;
		m_bInvertMouse = true;

		// Implement physics debug rendering functions.
		DemoPhysicsRender::Initialise(pApp);

		// Initialise mouse.
		m_pMouse = m_pApp->CreateMouse();
		m_pMouse->Initialise(true);

		// Initialise camera.
		m_pCamera = m_pApp->CreateCamera();
		m_pCamera->Initialise();

		// Initialise Physics.
		AABB worldAABB;
		worldAABB.Initialise(k_v3Zero, Vec3(5000.0f, 5000.0f, 5000.0f));
		Physics::CreateInstance(worldAABB);	
		Physics::GetInstance().SetGravity(-1.0f * 9.81f * k_v3UnitY);
		Physics::GetInstance().SetRenderCollisionsEnabled(false);


		//---------------------------------------------------------------------------------
		// Initialise Landscape
		//---------------------------------------------------------------------------------

		// Initialise Landscape collision object.
		m_pLandscape = new Landscape();
		m_pLandscape->Initialise(m_pApp, 0.0f);

		// Create a static object for the collision.
		StaticObject* pStaticObject = StaticObject::CreateNew();
		pStaticObject->Initialise(m_pLandscape->GetCollisionObject());

		// Add the static object to Physics.
		Physics::GetInstance().AddStaticObject(pStaticObject);
		// If we wanted to keep call Physics::GetInstance().RemoveStaticObject(pStaticObject) 
		// later, we would have to keep a pointer to pStaticObject.
		pStaticObject->Release(); 
		pStaticObject = 0;

		m_pRaceTrack = new RaceTrack();
		m_pRaceTrack->Initialise(m_pApp, "Assets/RaceTrack.txt");

				
		//---------------------------------------------------------------------------------
		// Create a car template.
		//---------------------------------------------------------------------------------
		CarTemplate* pCarTemplate = CarTemplate::CreateNew();
		pCarTemplate->BeginInitialise();

		// Make collision for car out of 2 convex collision objects.
		CollisionObjectConvex* pCollisionObjectConvex = CollisionObjectConvex::CreateNew();
		const int nPointList1Size = 8;
		const Vec3 pv3PointList1[nPointList1Size] = 
		{
			Vec3(-1.25f, -0.6f, -2.2f),
			Vec3(1.25f, -0.6f, -2.2f),
			Vec3(-1.25f, -0.6f, 2.2f),
			Vec3(1.25f, -0.6f, 2.2f),
			Vec3(-1.25f, 0.2f, -2.2f),
			Vec3(1.25f, 0.2f, -2.2f),
			Vec3(-1.25f, 0.2f, 2.2f),
			Vec3(1.25f, 0.2f, 2.2f),
		};
		pCollisionObjectConvex->InitialiseFromPointList(pv3PointList1, nPointList1Size);
		pCarTemplate->AddCollisionObject(pCollisionObjectConvex);
		pCollisionObjectConvex->Release();
		pCollisionObjectConvex = 0;
		
		const int nPointList2Size = 8;
		const Vec3 pv3PointList2[nPointList1Size] = 
		{
			Vec3(-1.25f, 0.2f, -1.3f),
			Vec3(1.25f, 0.2f, -1.3f),
			Vec3(-1.25f, 0.2f, 1.1f),
			Vec3(1.25f, 0.2f, 1.1f),

			Vec3(-1.0f, 0.6f, -1.2f),
			Vec3(1.0f, 0.6f, -1.2f),
			Vec3(-1.0f, 0.6f, 0.6f),
			Vec3(1.0f, 0.6f, 0.6f),
		};
		pCollisionObjectConvex = CollisionObjectConvex::CreateNew();
		pCollisionObjectConvex->InitialiseFromPointList(pv3PointList2, nPointList2Size);
		pCarTemplate->AddCollisionObject(pCollisionObjectConvex);
		pCollisionObjectConvex->Release();
		pCollisionObjectConvex = 0;
		

		// Add wheels
		const float k_fWheelRadius = 0.3f;
		const float k_fWheelWidth = 1.1f;
		const float k_fWheelTop = 0.0f;
		const float k_fWheelBottom = -1.0f;
		const float k_fWheelFront = 1.5f;
		const float k_fWheelBack = -1.5f;
		CarTemplate::Wheel wheel;


		// Front wheels
		wheel.SetSuspensionTop(Vec3(k_fWheelWidth, k_fWheelTop, k_fWheelFront));
		wheel.SetSuspensionBottom(Vec3(k_fWheelWidth, k_fWheelBottom, k_fWheelFront));
		wheel.SetRadius(k_fWheelRadius);
		wheel.SetSteeringFlag(true);
		wheel.SetDrivingFlag(false);
		wheel.SetHandBrakeFlag(false);
		wheel.SetSlipAngleToLateralForceCallBack(SlipAngleToLateralForce); // Use a non default slip curve
		wheel.SetSlipRatioToNormalisedTractionCallBack(SlipRatioToNormalisedTraction); // Use a non default slip curve
		pCarTemplate->AddWheel(wheel);

		wheel.SetSuspensionTop(Vec3(-k_fWheelWidth, k_fWheelTop, k_fWheelFront));
		wheel.SetSuspensionBottom(Vec3(-k_fWheelWidth, k_fWheelBottom, k_fWheelFront));
		wheel.SetRadius(k_fWheelRadius);
		wheel.SetSteeringFlag(true);
		wheel.SetDrivingFlag(false);
		wheel.SetHandBrakeFlag(false);
		pCarTemplate->AddWheel(wheel);

		// Rear wheels
		wheel.SetSuspensionTop(Vec3(k_fWheelWidth, k_fWheelTop, k_fWheelBack));
		wheel.SetSuspensionBottom(Vec3(k_fWheelWidth, k_fWheelBottom, k_fWheelBack));
		wheel.SetRadius(k_fWheelRadius);
		wheel.SetSteeringFlag(false);
		wheel.SetDrivingFlag(true);
		wheel.SetHandBrakeFlag(true);
		pCarTemplate->AddWheel(wheel);

		wheel.SetSuspensionTop(Vec3(-k_fWheelWidth, k_fWheelTop, k_fWheelBack));
		wheel.SetSuspensionBottom(Vec3(-k_fWheelWidth, k_fWheelBottom, k_fWheelBack));
		wheel.SetRadius(k_fWheelRadius);
		wheel.SetSteeringFlag(false);
		wheel.SetDrivingFlag(true);
		wheel.SetHandBrakeFlag(true);
		pCarTemplate->AddWheel(wheel);

		pCarTemplate->SetMass(1500.0f);
		pCarTemplate->SetTorqueMultiplier(5.0f);
		pCarTemplate->SetDownForce(4.0f);

		pCarTemplate->EndInitialise();
		
		// Initialise cars
		m_nNumDynamicObjects = 40;
		m_ppDynamicObjectList = new DynamicObjectCar* [m_nNumDynamicObjects];
		m_pRacingAIList = new RacingAI[m_nNumDynamicObjects];
		for (int nDynamicObject = 0; nDynamicObject < m_nNumDynamicObjects; nDynamicObject++)
		{
			int nPos = nDynamicObject;
			if (nPos == 0)
				nPos = 30;
			else if (nPos == 30)
				nPos = 0;

			MFrame startFrame = m_pRaceTrack->GetStartFrame(nPos);
			Vec3 v3GroundPos = startFrame.v3Translation;
			v3GroundPos.x += 0.01f;
			v3GroundPos.z += 0.01f;

			// Get placement height
			Collision collision = Physics::GetInstance().TestLineForCollision(
				v3GroundPos + k_v3UnitY * 1000.0f,
				v3GroundPos -k_v3UnitY * 2000.0f,
				Physics::FLAG_ALL_NON_DYNAMIC_OBJECTS);

			if (collision.CollisionOccurred())
				v3GroundPos = collision.GetPosition();
			v3GroundPos.y += 1.0f;

			m_ppDynamicObjectList[nDynamicObject] = DynamicObjectCar::CreateNew();
			TA_ASSERT(m_ppDynamicObjectList[nDynamicObject]);
			DynamicObjectCar& dynamicObject = *m_ppDynamicObjectList[nDynamicObject];
			dynamicObject.Initialise(pCarTemplate);
			dynamicObject.SetFrame(
				MFrame(
					v3GroundPos,
					Mat33(Mat33::ROTATION_AXIS_Y, k_fPi * 0.0f * nDynamicObject)));
			//v3GroundPos.y += 2.5f;
			v3GroundPos.y += 0.98f;
			dynamicObject.SetFriction(0.5f);

			dynamicObject.SetCenterOffset(dynamicObject.GetCenterOffset() + Vec3(0.0f, -0.3f, 0.0f));
			dynamicObject.SetAntiRollFudge(0.28f);
			Physics::GetInstance().AddDynamicObject(&dynamicObject);	

			float fSpeed = 1.0f;
			if (nDynamicObject != 0)
			{
				fSpeed = 
					(((float)(m_nNumDynamicObjects - nPos + 1) / (float)m_nNumDynamicObjects) + 1.0f) * 0.5f + 
					RandFunc::RandBetween(0.0f, 0.1f);
			}
			m_pRacingAIList[nDynamicObject].Initialise(dynamicObject, fSpeed);
		}
		
		pCarTemplate->Release();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Finalise()
	{
		if (m_ppDynamicObjectList)
		{
			for (int i = 0; i < m_nNumDynamicObjects; i++)
			{
				if (m_ppDynamicObjectList[i])
				{
					m_ppDynamicObjectList[i]->Release();
					m_ppDynamicObjectList[i] = 0;
				}
			}
			delete []m_ppDynamicObjectList;
			m_ppDynamicObjectList = 0;
		}

		if (m_pRacingAIList)
		{
			delete []m_pRacingAIList;
			m_pRacingAIList = 0;
		}

		if (m_pRaceTrack)
		{
			delete m_pRaceTrack;
			m_pRaceTrack = 0;
		}
		if (m_pLandscape)
		{
			delete m_pLandscape;
			m_pLandscape = 0;
		}
		if (m_pMouse)
		{
			m_pMouse->Release();
			m_pMouse = 0;
		}
		if (m_pCamera)
		{
			m_pCamera->Release();
			m_pCamera = 0;
		}
		m_pApp = 0;

		Physics::DestroyInstance();

		DemoPhysicsRender::Finalise();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Update(float fDt)
	{
		TA_ASSERT(m_ppDynamicObjectList);

		// Update Input
		m_pMouse->Update();

		if (KeyIsJustPressed(VK_F1))
			m_bHelp = !m_bHelp;
		if (KeyIsJustPressed('I'))
			m_bInvertMouse = !m_bInvertMouse;

		{
			const float k_fSteeringSensitivity = 2.5f;
			float fSteering = 0.0f;
			if (KeyIsDown(VK_NUMPAD4) || KeyIsDown(VK_LEFT) || KeyIsDown('A'))
				fSteering -= 1.0f;
			if (KeyIsDown(VK_NUMPAD6) || KeyIsDown(VK_RIGHT) || KeyIsDown('D'))
				fSteering += 1.0f;
			static float s_fSteering = 0.0f;
			if (fSteering > s_fSteering)
			{
				s_fSteering += fDt * k_fSteeringSensitivity;
				if (s_fSteering > fSteering)
					s_fSteering = fSteering;
			}
			else if (fSteering < s_fSteering)
			{
				s_fSteering -= fDt * k_fSteeringSensitivity;
				if (s_fSteering < fSteering)
					s_fSteering = fSteering;
			}

	
			float fSteeringHelp = KeyIsDown('Z') ? 0.0f : 1.0f;
			Clamp(s_fSteering, -1.0f, 1.0f);
			if (m_ppDynamicObjectList[0]->GetLinearVelocity().Dot(
					m_ppDynamicObjectList[0]->GetFrame().m33Rotation.v3Z) > 0.0f)
			{
				/*s_fSteering -= 
					fSteeringHelp *
					m_ppDynamicObjectList[0]->GetAngularVelocity().y * 
					(0.004f + 0.01f * (1.0f - Fabs(s_fSteering))) ;
				Clamp(s_fSteering, -1.0f, 1.0f);*/
			}

			float fTractionControl = KeyIsDown('X') ? 0.0f : 1.0f;

			float fAcceleration = (KeyIsDown(VK_NUMPAD8) || KeyIsDown(VK_UP) || KeyIsDown('W')) ? 1.0f : 0.0f;
			float fBrake = (KeyIsDown(VK_NUMPAD5) || KeyIsDown(VK_DOWN) || KeyIsDown('S')) ? 1.0f : 0.0f;
			float fHandBrake = KeyIsDown(VK_SPACE) ? 1.0f : 0.0f;


			TA_ASSERT(m_ppDynamicObjectList[0]);
			const int nGear = -1; // automatic gear changes
			// manual gear changes
			/*static int nGear = 1;
			if (KeyIsJustPressed('R'))
				nGear++;
			if (KeyIsJustPressed('F'))
				nGear--;
			Clamp(nGear, 0, 6);*/
			m_ppDynamicObjectList[0]->UpdateControls(
				s_fSteering,
				fAcceleration,
				fBrake,
				fHandBrake,
				fSteeringHelp,
				fTractionControl, 
				nGear);
			// Make sure that the players car does not fall asleep.
			Physics::GetInstance().SetDynamicObjectIsMoving(m_ppDynamicObjectList[0]);
		}
		
		for (int nDynamicObject = 1; nDynamicObject < m_nNumDynamicObjects; nDynamicObject++)			
			m_pRacingAIList[nDynamicObject].Update(fDt, m_pRaceTrack);

		// Update physics
		Physics::GetInstance().Update(fDt);

		UpdateCamera(fDt);
	}

	void UpdateCamera(float fDt)
	{
		// Implemented a camera to follow the car.

		m_pCamera->SetNearClip(0.1f);
		m_pCamera->SetFarClip(1000.0f);		

		Vec3 v3Position = m_ppDynamicObjectList[0]->GetPosition();

		const Vec3 k_v3fLookAtPos(0.0f, 1.0f, 0.0f);
		const Vec3 k_v3LookFromPos(0.0f, 3.0f, 0.0f);
		float fViewDistance = 10.0f;
		const float k_fCarSettleTime = 2.0f;

		// Using statics here because of lazy programming
		static Vec3 s_v3LastBaseLookFromPos = 
			m_ppDynamicObjectList[0]->GetPosition() - 
			m_ppDynamicObjectList[0]->GetFrame().m33Rotation.v3Z * 20.0f;
		static float s_fFacing = 1.0f;
		static float s_fCarSettleTime = 0.0f;

		// camera intro
		if (s_fCarSettleTime < k_fCarSettleTime)
		{
			Vec3 v3DefaultPosition = 
				m_ppDynamicObjectList[0]->GetPosition() - 
				m_ppDynamicObjectList[0]->GetFrame().m33Rotation.v3Z * 20.0f;
			if (s_fCarSettleTime > k_fCarSettleTime - 1.0f)
				s_v3LastBaseLookFromPos += (k_fCarSettleTime - s_fCarSettleTime) * (v3DefaultPosition - s_v3LastBaseLookFromPos);
			else
				s_v3LastBaseLookFromPos = v3DefaultPosition;
			fViewDistance += (k_fCarSettleTime - s_fCarSettleTime) * (k_fCarSettleTime - s_fCarSettleTime);				
			s_fCarSettleTime += fDt;
		}

		Vec3 v3LookFromPos;

		// First we nudge the previous camera position to get it to behave
		// exactly the way we want.

		// if the object is moving towards the camera then nudge the camera
		// side was a little so it goes around.
		// This avoids camera going over the top of the car and flipping when 
		// changing velocity when going into reverse while looking forwards.
		float fNudge = (s_v3LastBaseLookFromPos - v3Position).GetMagnitude();
		if (fNudge < fViewDistance)
		{
			fNudge = fViewDistance - fNudge;
			s_v3LastBaseLookFromPos -= m_ppDynamicObjectList[0]->GetFrame().m33Rotation.v3X * fNudge * 1.1f;
			s_v3LastBaseLookFromPos.y += fNudge * 0.2f;
		}

		// Nudged the camera a bit so that it swings around a bit more to face
		// the direction that the car is moving.
		// This makes it a little easier to see around corners.
		s_v3LastBaseLookFromPos -= m_ppDynamicObjectList[0]->GetLinearVelocity() * 0.1f;

		// Nudge the camera so that is a bit more likely to look in the direction
		// that the car is facing. This helps most when maneuvering at low speed.
		Vec3 v3FacingNudge = m_ppDynamicObjectList[0]->GetFrame().m33Rotation.v3Z * 0.2f * s_fFacing;
		v3FacingNudge.y = 0.0f;
		s_v3LastBaseLookFromPos -= v3FacingNudge;
		// Smoothly choose whether the car is facing forwards or backwards for the above code.
		float fDot = (v3Position - s_v3LastBaseLookFromPos).Dot(m_ppDynamicObjectList[0]->GetFrame().m33Rotation.v3Z);
		s_fFacing = fDot * 0.1f;
		ClampPosNeg(s_fFacing, 1.0f);

		// Now we make the camera follow the new object position.

		Vec3 v3LookDirection = v3Position - s_v3LastBaseLookFromPos;
		float fMag = v3LookDirection.GetMagnitude();
		if (fMag > 0.0f)
			v3LookDirection /= fMag;
		else
			v3LookDirection = k_v3UnitZ; // It would be much better to use the last valid value

		v3LookFromPos = v3Position - v3LookDirection * fViewDistance;

		s_v3LastBaseLookFromPos = v3LookFromPos;

		if (Fabs(v3LookDirection.y) <= 0.99f)
		{			
			Mat33 m33;
			m33.SetToLookAt(v3LookDirection, k_v3UnitY);

			v3LookFromPos += k_v3LookFromPos * m33;

			Vec3 v3LookAtPos = v3Position + k_v3fLookAtPos * m33;

			m_pCamera->SetPosition(v3LookFromPos);

			Mat33 m33CameraLookAt;
			m33CameraLookAt.SetToLookAt((v3LookAtPos - v3LookFromPos).GetNormal(), m33.v3Y);

			m_pCamera->SetRotation(m33CameraLookAt);	
		}

	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Render()
	{
		// Set the camera
		m_pCamera->Realise();
		
		// Render physics (renders collision objects)
		Physics::GetInstance().Render();
		
		for (int nDynamicObject = 0; nDynamicObject < m_nNumDynamicObjects; nDynamicObject++)
		{
			DynamicObjectCar& car = *m_ppDynamicObjectList[nDynamicObject];
			if (!car.AddedToPhysics())
				continue; // car has been removed from physics in racingAI so dont render wheels
			for (int nWheel = 0; nWheel < car.GetNumWheels(); nWheel++)
			{
				const DynamicObjectCar::Wheel& wheel = car.GetWheel(nWheel);
				int nSize = 8;
				for (int i = 0; i < nSize; i++)
				{
					float fSinA;
					float fCosA;
					SinAndCos(2.0f * k_fPi * (float)i / (float)nSize, fSinA, fCosA);
					Vec3 v3PosA = 
						Vec3(0.0f, fSinA, fCosA) * wheel.GetRadius() * 
						wheel.GetLocalGraphicsFrame() * 
						car.GetGraphicsFrame();
					float fSinB;
					float fCosB;
					SinAndCos(2.0f * k_fPi * (float)(i + 1) / (float)nSize, fSinB, fCosB);
					Vec3 v3PosB = 
						Vec3(0.0f, fSinB, fCosB) * wheel.GetRadius() * 
						wheel.GetLocalGraphicsFrame() *
						car.GetGraphicsFrame();
					m_pApp->RenderLine(v3PosA, v3PosB, 0xFFFFFFFF);
				}
			}
		}

		m_pApp->RenderText(
			"KPH: %3.0f Gear: %d RPM: %4.0f", 
			m_ppDynamicObjectList[0]->GetCarSpeedFromWheels() * 3.6f,
			m_ppDynamicObjectList[0]->GetGear(),
			m_ppDynamicObjectList[0]->GetRPM());


		// Render the Landscape
		m_pLandscape->Render();
		m_pRaceTrack->Render();

		RenderHelp();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void RenderHelp()
	{
		if (m_bHelp)
		{
			m_pApp->RenderText("Toggle help on and off (F1)");
			m_pApp->RenderText("Forward (Up Arrow)");
			m_pApp->RenderText("Backward (Down Arrow)");
			m_pApp->RenderText("Handbrake (Space)");
			m_pApp->RenderText("Disable steering help (Z)");
			m_pApp->RenderText("Disable traction control (X)");
			m_pApp->RenderText("Quit (Esc)");
		}
	}

private:	
	IApp* m_pApp;
	IMouse* m_pMouse;	
	ICamera* m_pCamera;
	Landscape* m_pLandscape;
	RaceTrack* m_pRaceTrack;

	int m_nNumDynamicObjects;
	DynamicObjectCar** m_ppDynamicObjectList;
	RacingAI* m_pRacingAIList;
	bool m_bHelp;
	bool m_bInvertMouse;

	
	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	static float TAC_CALL SlipAngleToLateralForce(
		float fSlipAngle, 
		float fLongitudinalForce, 
		const SurfaceAttribute& surfaceAttribute)
	{
		const float fInitialGradient = 14.0f;
		const float fEndGradient = -0.1f;
		const float fEndOffset = 2.0f;

		const float fSegmentEndA = 0.08f;
		const float fSegmentEndB = 0.27f;

		const float fSegmentEndAOut = fInitialGradient * fSegmentEndA;
		const float fSegmentEndBOut = fEndGradient * fSegmentEndB + fEndOffset;

		const float fInvSegBLength = 1.0f / (fSegmentEndB - fSegmentEndA);
		const float fCubicGradA = fInitialGradient * (fSegmentEndB - fSegmentEndA);
		const float fCubicGradB = fEndGradient * (fSegmentEndB - fSegmentEndA);

		float fSign = Sign(fSlipAngle);
		fSlipAngle *= fSign;

		float fResult;
		if (fSlipAngle < fSegmentEndA)
		{
			fResult = fSign * fInitialGradient * fSlipAngle;
		}
		else if (fSlipAngle < fSegmentEndB)
		{		
			fResult = fSign * Geometry::CubicInterpolate(
				fSegmentEndAOut, 
				fSegmentEndBOut,
				fCubicGradA, 
				fCubicGradB, 
				(fSlipAngle - fSegmentEndA) * fInvSegBLength);
		}
		else
		{
			float fValue = fEndGradient * fSlipAngle + fEndOffset;
			if (fValue < 0.0f)
				fValue = 0.0f;
			fResult = fSign * fValue;
		}

		fResult *= 2.5f;
		fResult /= (Fabs(fLongitudinalForce) * 8.5f + 1.0f);
		return fResult;
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	static float TAC_CALL SlipRatioToNormalisedTraction(
		float fSlipRatio, 
		const SurfaceAttribute& surfaceAttribute)
	{
		const float fInitialGradient = 7.0f;
		const float fEndGradient = -0.2f;
		const float fEndOffset = 1.2f;

		const float fSegmentEndA = 0.12f;
		const float fSegmentEndB = 0.25f;

		const float fSegmentEndAOut = fInitialGradient * fSegmentEndA;
		const float fSegmentEndBOut = fEndGradient * fSegmentEndB + fEndOffset;

		const float fInvSegBLength = 1.0f / (fSegmentEndB - fSegmentEndA);
		const float fCubicGradA = fInitialGradient * (fSegmentEndB - fSegmentEndA);
		const float fCubicGradB = fEndGradient * (fSegmentEndB - fSegmentEndA);
		
		float fSign = Sign(fSlipRatio);
		fSlipRatio *= fSign;

		fSign *= 8.0f;

		if (fSlipRatio < fSegmentEndA)
		{
			return fSign * fInitialGradient * fSlipRatio;
		}
		else if (fSlipRatio < fSegmentEndB)
		{		
			return fSign * Geometry::CubicInterpolate(
				fSegmentEndAOut, 
				fSegmentEndBOut,
				fCubicGradA, 
				fCubicGradB, 
				(fSlipRatio - fSegmentEndA) * fInvSegBLength);
		}
		else
		{
			float fValue = fEndGradient * fSlipRatio + fEndOffset;
			if (fValue < 0.0f)
				fValue = 0.0f;
			return fSign * fValue;
		}
	}
};

TA_IMPLEMENT_DEMO_PLUGIN(DemoCars);