//---------------------------------------------------------------------------------
// File Name: DemoRagDoll.cpp
// Description:
//
// Copyright (C) 2004 - 2006 True Axis Pty Ltd, Australia. 
// All Rights Reserved.
//
// Use of this code is covered by the "TRUE AXIS PHYSICS SDK DEMOS SOURCE LICENSE AGREEMENT".
// Details of this agreement are found in the file "TrueAxisPhysicsSDKDemosSourceLicence.txt".
// You have permission to modify this file and use it for any purpose under the terms of
// the agreement.
//
// History:
//		Created File.
//---------------------------------------------------------------------------------

#include "../Interface/IApp.h"
#include "../Interface/IDemoPlugin.h"
#include "../Interface/IMouse.h"
#include "../Interface/ICamera.h"
#include "../Interface/IStaticGeometry.h"

#include "../TA/Common/RandFunc.h"

#include "../TA/Physics/Physics.h"
#include "../TA/Physics/CollisionObjectAABBMesh.h"
#include "../TA/Physics/CollisionObjectCombo.h"

#include "../Shared/DemoPhysicsRender.h"

#include "Ragdoll.h"

#include <windows.h>

using namespace TA;

static StaticObject* pStaticObject = 0;
	
static bool KeyIsDown(int nKey) { return (GetAsyncKeyState(nKey) & 0x8000) != 0; }
static bool KeyIsJustPressed(int nKey) { return (GetAsyncKeyState(nKey) & 0x8001) == 0x8001; }

class DemoRagDoll : public IDemoPlugin
{
public:
	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	DemoRagDoll()
	{
		m_nMaxNumDynamicObjects = 0;
		m_ppDynamicObjectList = 0;
		m_nMaxNumRagDolls = 0;
		m_nNextRagDoll = 0;
		m_ppRagDollList = 0;
		m_v3CameraPosition.Clear();
		m_fImpulse = 0.0f;
		m_fAngularImpulse = 0.0f;
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	~DemoRagDoll()
	{
		Finalise();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Release() { delete this; }

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Initialise(IApp* pApp)
	{
		m_pApp = pApp;

		// Implement physics debug rendering functions.
		DemoPhysicsRender::Initialise(pApp);

		// Initialise mouse.
		m_pMouse = m_pApp->CreateMouse();
		m_pMouse->Initialise(true);

		// Initialise camera.
		m_pCamera = m_pApp->CreateCamera();
		m_pCamera->Initialise();

		m_fImpulse = 0.0f;
		m_fAngularImpulse = 0.0f;

		// Initialise Physics.
		AABB worldAABB;
		worldAABB.Initialise(
			k_v3Zero, 
			Vec3(500.0f, 500.0f, 500.0f));
		Physics::CreateInstance(
			worldAABB,
			Vec3(0.0f, -9.81f, 0.0f),
			Physics::FLAG_XZ_COLLISION_GRID,
			4.0f);

		Physics::GetInstance().SetMaxNumCollisions(2048);

		//---------------------------------------------------------------------------------
		// Initialise Static Geometry
		//---------------------------------------------------------------------------------
		m_pStaticGeometry = m_pApp->CreateStaticGeometry();
		m_pStaticGeometry->Initialise("Assets/Data/TestScene.X", 0.004f);

		if (m_pStaticGeometry->Lock())
		{
			CollisionObjectAABBMesh* pCollisionObject = CollisionObjectAABBMesh::CreateNew();
			pCollisionObject->Initialise(
				m_pStaticGeometry->GetNumVertices(),
				m_pStaticGeometry->GetNumTriangles(),
				m_pStaticGeometry->GetNumTriangles() * 3);

			for (int nVertex = 0; nVertex < m_pStaticGeometry->GetNumVertices(); nVertex++)
				pCollisionObject->AddVertex(m_pStaticGeometry->GetVertexPosition(nVertex));

			for (int nTriangle = 0; nTriangle < m_pStaticGeometry->GetNumTriangles(); nTriangle++)
			{
				int pnIndexList[3] = 
				{
					m_pStaticGeometry->GetTriangleVertexIndex(nTriangle, 0),
					m_pStaticGeometry->GetTriangleVertexIndex(nTriangle, 2),
					m_pStaticGeometry->GetTriangleVertexIndex(nTriangle, 1),
				};
				pCollisionObject->AddPolygon(3, pnIndexList);
			}

			pCollisionObject->FinishedAddingGeometry();
			m_pStaticGeometry->Unlock();

			// Create a static object for the collision.
			pStaticObject = StaticObject::CreateNew();
			pStaticObject->Initialise(pCollisionObject);
			pCollisionObject->Release();

			// Add the static object to Physics.
			Physics::GetInstance().AddStaticObject(pStaticObject);
			pStaticObject->Release();		
		}



		//---------------------------------------------------------------------------------
		// Initialise Dynamic Objects
		//---------------------------------------------------------------------------------

		// Initialise collision object
		CollisionObjectCombo* pCollisionObjectCombo = CollisionObjectCombo::CreateNew();
		pCollisionObjectCombo->InitialiseAsABox(1.0f, 1.0f, 1.0f);

		Vec3 v3GroundPos(0.0f, 0.0f, 0.0f);
		// Get placement height
		Collision collision = Physics::GetInstance().TestLineForCollision(
			v3GroundPos + k_v3UnitY * 1000.0f,
			-k_v3UnitY * 2000.0f,
			Physics::FLAG_ALL_NON_DYNAMIC_OBJECTS);

		if (collision.CollisionOccurred())
			v3GroundPos = collision.GetPosition();
		v3GroundPos.y += 5.0f;

		m_v3CameraPosition = v3GroundPos;
		
		
		v3GroundPos.Initialise(-5.0f, 0.0f, 7.0f);
		// Get placement height
		collision = Physics::GetInstance().TestLineForCollision(
			v3GroundPos + k_v3UnitY * 1000.0f,
			-k_v3UnitY * 2000.0f,
			Physics::FLAG_ALL_NON_DYNAMIC_OBJECTS);
		if (collision.CollisionOccurred())
			v3GroundPos = collision.GetPosition();
		v3GroundPos.y += 0.5f;
		
		// Initialise objects
		m_nNumDynamicObjects = 3;
		m_nMaxNumDynamicObjects = 3;
		m_ppDynamicObjectList = new DynamicObject* [m_nMaxNumDynamicObjects];
		for (int nDynamicObject = 0; nDynamicObject < m_nMaxNumDynamicObjects; nDynamicObject++)
		{
			m_ppDynamicObjectList[nDynamicObject] = DynamicObject::CreateNew();
			TA_ASSERT(m_ppDynamicObjectList[nDynamicObject]);
			DynamicObject& dynamicObject = *m_ppDynamicObjectList[nDynamicObject];
			dynamicObject.Initialise(pCollisionObjectCombo);
			dynamicObject.SetFrame(MFrame(v3GroundPos));
			v3GroundPos.y += 0.99f;		
			Physics::GetInstance().AddDynamicObject(&dynamicObject, false);
		}
		pCollisionObjectCombo->Release();

        m_nMaxNumRagDolls = 100;
		m_nNextRagDoll = 0;
		m_ppRagDollList = new RagDoll* [m_nMaxNumRagDolls];
		for (int nRagDoll = 0; nRagDoll < m_nMaxNumRagDolls; nRagDoll++)
		{
			m_ppRagDollList[nRagDoll] = new RagDoll();
			TA_ASSERT(m_ppRagDollList[nRagDoll]);
			m_ppRagDollList[nRagDoll]->Initialise();
		}
		
		m_bHelp = true;
		m_bInvertMouse = true;	
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Reset()
	{
		for (int nRagDoll = 0; nRagDoll < m_nMaxNumRagDolls; nRagDoll++)
			m_ppRagDollList[nRagDoll]->RemoveFromPhysics();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Finalise()
	{
		if (m_ppDynamicObjectList)
		{
			for (int i = 0; i < m_nMaxNumDynamicObjects; i++)
			{
				if (m_ppDynamicObjectList[i])
				{
					m_ppDynamicObjectList[i]->Release();
					m_ppDynamicObjectList[i] = 0;
				}
			}
			delete []m_ppDynamicObjectList;
			m_ppDynamicObjectList = 0;
		}
		if (m_ppRagDollList)
		{
			for (int i = 0; i < m_nMaxNumRagDolls; i++)
			{
				if (m_ppRagDollList[i])
				{
					delete m_ppRagDollList[i];
					m_ppRagDollList[i] = 0;
				}
			}
			delete []m_ppRagDollList;
			m_ppRagDollList = 0;
		}
		m_nNextRagDoll = 0;
	
		m_v3CameraPosition.Clear();
		m_fImpulse = 0.0f;
		m_fAngularImpulse = 0.0f;

		if (m_pStaticGeometry)
		{
			m_pStaticGeometry->Release();
			m_pStaticGeometry = 0;
		}
		if (m_pMouse)
		{
			m_pMouse->Release();
			m_pMouse = 0;
		}
		if (m_pCamera)
		{
			m_pCamera->Release();
			m_pCamera = 0;
		}
		m_pApp = 0;

		Physics::DestroyInstance();

		DemoPhysicsRender::Finalise();

	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Update(float fDt)
	{
		fDt = 1.0f / 60.0f;
		// Update Input
		m_pMouse->Update();

		if (m_pMouse->IsLeftButtonDown() || m_pMouse->IsRightButtonDown())
		{
			if (m_pMouse->IsLeftButtonDown())
				m_fImpulse += fDt * 50.0f;

			if (m_pMouse->IsRightButtonDown())
				m_fAngularImpulse += fDt * 50.0f;
		}
		else if (m_fImpulse > 0.0f || m_fAngularImpulse > 0.0f)
		{		
			if (m_fImpulse > 400)
				m_fImpulse = 400.0f;
			TA_ASSERT(m_ppRagDollList[m_nNextRagDoll]);
			RagDoll& ragDoll = *m_ppRagDollList[m_nNextRagDoll];
			ragDoll.AddToPhysics();

			ragDoll.SetFrame(MFrame(m_pCamera->GetPosition() +  Vec3(0.0f, 0.0f, 0.0f)));

			ragDoll.ApplyTestImpulse(
				m_pCamera->GetRotation().v3Z * 20.0f * m_fImpulse +
				m_pCamera->GetRotation().v3Y * 5.0f * m_fImpulse,
				ragDoll.GetPosition());
			
			ragDoll.ApplyTestAngularImpulse(k_v3UnitY * m_fAngularImpulse);

			m_fImpulse = 0.0f;
			m_fAngularImpulse = 0.0f;
			m_nNextRagDoll++;
			if (m_nNextRagDoll >= m_nMaxNumRagDolls)
				m_nNextRagDoll = 0;
		}

		if (KeyIsJustPressed(VK_F1))
			m_bHelp = !m_bHelp;
		if (KeyIsJustPressed('I'))
			m_bInvertMouse = !m_bInvertMouse;
		if (KeyIsJustPressed('R'))
			Reset();
	
		// Update physics
		Physics::GetInstance().Update(fDt);

		if (m_ppRagDollList)
		{
			for (int i = 0; i < m_nMaxNumRagDolls; i++)
			{
				if (m_ppRagDollList[i])
				{
					m_ppRagDollList[i]->Update(fDt);
				}
			}
		}


		// Update the camera.	
		m_pCamera->SetNearClip(0.1f);
		m_pCamera->SetFarClip(1000.0f);
		m_pCamera->SetViewDistance(0.5f);
		m_pCamera->UpdateRotation(
			m_pMouse->GetXMovement(),
			m_bInvertMouse ? m_pMouse->GetYMovement() : -m_pMouse->GetYMovement());
		m_pCamera->SetTargetPosition(m_v3CameraPosition);
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Render()
	{
		// Set the camera
		m_pCamera->Realise();
		
		// Render physics (renders collision objects)
		Physics::GetInstance().Render();

		// Render the StaticGeometry
		m_pStaticGeometry->Render();

		if (m_ppRagDollList)
		{
			for (int i = 0; i < m_nMaxNumRagDolls; i++)
			{
				if (m_ppRagDollList[i])
					m_ppRagDollList[i]->Render();
			}
		}

		RenderHelp();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void RenderHelp()
	{
		if (m_bHelp)
		{
			m_pApp->RenderText("Toggle help on and off (F1)");
			m_pApp->RenderText("Move the mouse to rotate the camera");
			m_pApp->RenderText("Invert mouse (I)");
			m_pApp->RenderText("(Left Mouse Button) Hold to build up power, release to throw object");
			m_pApp->RenderText("Remove all ragdolls (R)");
			m_pApp->RenderText("Quit (Esc)");
		}
	}

private:	
	IApp* m_pApp;
	IMouse* m_pMouse;	
	ICamera* m_pCamera;
	IStaticGeometry* m_pStaticGeometry;

	int m_nMaxNumDynamicObjects;
	int m_nNumDynamicObjects;
	DynamicObject** m_ppDynamicObjectList;
	int m_nMaxNumRagDolls;
	int m_nNextRagDoll;
	RagDoll** m_ppRagDollList;

	float m_fTimeUntilNextBoxAppear;
	Vec3 m_v3CameraPosition;
	float m_fImpulse;
	float m_fAngularImpulse;
	bool m_bHelp;
	bool m_bInvertMouse;

};

TA_IMPLEMENT_DEMO_PLUGIN(DemoRagDoll);