//---------------------------------------------------------------------------------
// File Name: DemoStackingObjects.cpp
// Description: 	
//
// Copyright (C) 2004 - 2006 True Axis Pty Ltd, Australia. 
// All Rights Reserved.
//
// Use of this code is covered by the "TRUE AXIS PHYSICS SDK DEMOS SOURCE LICENSE AGREEMENT".
// Details of this agreement are found in the file "TrueAxisPhysicsSDKDemosSourceLicence.txt".
// You have permission to modify this file and use it for any purpose under the terms of
// the agreement.
//
// History:
//		Created File.
//---------------------------------------------------------------------------------

#include "../Interface/IApp.h"
#include "../Interface/IDemoPlugin.h"
#include "../Interface/IMouse.h"
#include "../Interface/ICamera.h"

#include "../TA/Common/RandFunc.h"

#include "../TA/Physics/Physics.h"
#include "../TA/Physics/CollisionObjectAABBMesh.h"
#include "../TA/Physics/CollisionObjectCombo.h"

#include "../Shared/Landscape.h"
#include "../Shared/DemoPhysicsRender.h"

#include <windows.h>

using namespace TA;
	
static bool KeyIsDown(int nKey) { return (GetAsyncKeyState(nKey) & 0x8000) != 0; }
static bool KeyIsJustPressed(int nKey) { return (GetAsyncKeyState(nKey) & 0x8001) == 0x8001; }

static bool g_bRenderPhysics = true;

float k_fMassMult = 100.0f;

class DemoStackingObjects : public IDemoPlugin
{
public:

	enum ObjectType
	{
		COLLISION_BOX = 0,
		COLLISION_BRICK,
		COLLISION_BIG_BALL,
		COLLISION_LONG_BOX,
		COLLISION_LONG_CAPSULE,
		COLLISION_CYLINDER,
		COLLISION_CAPSULE,
		COLLISION_LONG_BOX2,
		NUM_COLLISION_OBJECTS
	};

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	DemoStackingObjects()
	{
		m_nMaxNumDynamicObjects = 0;
		m_nNumDynamicObjects = 0;
		m_ppDynamicObjectList = 0;
		m_v3CameraTarget.Clear();
		m_nMode = 0;
		m_fTimer = 0.0f;
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	~DemoStackingObjects()
	{
		Finalise();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Release()
	{
		delete this;
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Initialise(IApp* pApp)
	{
		m_pApp = pApp;

		// Implement physics debug rendering functions.
		DemoPhysicsRender::Initialise(pApp);

		// Initialise mouse.
		m_pMouse = m_pApp->CreateMouse();
		m_pMouse->Initialise(true);

		// Initialise camera.
		m_pCamera = m_pApp->CreateCamera();
		m_pCamera->Initialise();
		m_pCamera->SetViewDistance(10.0f);

		// Initialise Physics.
		AABB worldAABB;
		worldAABB.Initialise(k_v3Zero, Vec3(1000.0f, 1000.0f, 1000.0f));
		Physics::CreateInstance();
		Physics& physics = Physics::GetInstance();
		physics.SetWorldDimensions(worldAABB);
		physics.SetSpacialDivisionType(Physics::SPACIAL_DIVISION_TYPE_DYNAMIC_OCTREE);
		physics.SetSpacialDivisionMinSize(1.0f);
		physics.SetupSimulation();




		//---------------------------------------------------------------------------------
		// Initialise Landscape
		//---------------------------------------------------------------------------------

		// Initialise Landscape collision object.
		m_pLandscape = new Landscape();
		m_pLandscape->Initialise(m_pApp, 0.0f);

		// Create a static object for the collision.
		StaticObject* pStaticObject = StaticObject::CreateNew();
		pStaticObject->Initialise(m_pLandscape->GetCollisionObject());

		// Add the static object to Physics.
		Physics::GetInstance().AddStaticObject(pStaticObject);
		// If we wanted to keep call Physics::GetInstance().RemoveStaticObject(pStaticObject) 
		// later, we would have to keep a pointer to pStaticObject.
		pStaticObject->Release(); 
		pStaticObject = 0;



		//---------------------------------------------------------------------------------
		// Initialise Dynamic Objects
		//---------------------------------------------------------------------------------

		// Initialise collision object
		m_ppCollisionObjectCombo = new CollisionObjectCombo*[NUM_COLLISION_OBJECTS];
		m_ppCollisionObjectCombo[COLLISION_BOX] = CollisionObjectCombo::CreateNew();
		m_ppCollisionObjectCombo[COLLISION_BOX]->InitialiseAsABox(1.0f, 1.0f, 1.0f);
		m_ppCollisionObjectCombo[COLLISION_BRICK] = CollisionObjectCombo::CreateNew();
		m_ppCollisionObjectCombo[COLLISION_BRICK]->InitialiseAsABox(1.5f, 0.8f, 1.0f);
		m_ppCollisionObjectCombo[COLLISION_BIG_BALL] = CollisionObjectCombo::CreateNew();
		m_ppCollisionObjectCombo[COLLISION_BIG_BALL]->InitialiseAsASphere(Vec3(0.0f, 0.0f, 0.0f), 2.0f);
		m_ppCollisionObjectCombo[COLLISION_LONG_BOX] = CollisionObjectCombo::CreateNew();
		m_ppCollisionObjectCombo[COLLISION_LONG_BOX]->InitialiseAsABox(6.5f, 0.4f, 1.0f);
		m_ppCollisionObjectCombo[COLLISION_LONG_CAPSULE] = CollisionObjectCombo::CreateNew();
		m_ppCollisionObjectCombo[COLLISION_LONG_CAPSULE]->InitialiseAsACapsule(Vec3(-3.5f, 0.0f, 0.0f), Vec3(3.5f, 0.0f, 0.0f), 0.25f);
		m_ppCollisionObjectCombo[COLLISION_CYLINDER] = CollisionObjectCombo::CreateNew();
		m_ppCollisionObjectCombo[COLLISION_CYLINDER]->InitialiseAsACylinder(Vec3(0.0f, 0.5f, 0.0f), Vec3(0.0f, -0.5f, 0.0f), 0.5f);
		m_ppCollisionObjectCombo[COLLISION_CAPSULE] = CollisionObjectCombo::CreateNew();
		m_ppCollisionObjectCombo[COLLISION_CAPSULE]->InitialiseAsACapsule(Vec3(-1.5f, 0.0f, 0.0f), Vec3(1.5f, 0.0f, 0.0f), 0.25f);
		m_ppCollisionObjectCombo[COLLISION_LONG_BOX2] = CollisionObjectCombo::CreateNew();
		m_ppCollisionObjectCombo[COLLISION_LONG_BOX2]->InitialiseAsABox(1.5f, 0.4f, 1.0f);

		

		
		
	
		Vec3 v3GroundPos(1.0f, 0.0f, 1.0f);
		// Get placement height
		Collision collision = Physics::GetInstance().TestLineForCollision(
			v3GroundPos + k_v3UnitY * 1000.0f,
			-k_v3UnitY * 2000.0f,
			Physics::FLAG_ALL_NON_DYNAMIC_OBJECTS);

		if (collision.CollisionOccurred())
			v3GroundPos = collision.GetPosition();
		v3GroundPos.y += 1.0f;
			
		m_v3CameraTarget = v3GroundPos;
		m_v3CameraTarget.y += 1.5f;

		m_nNumDynamicObjects = 0;
		m_nMaxNumDynamicObjects = 5000;
		m_ppDynamicObjectList = new DynamicObject* [m_nMaxNumDynamicObjects];
		for (int nDynamicObject = 0; nDynamicObject < m_nMaxNumDynamicObjects; nDynamicObject++)
		{
			m_ppDynamicObjectList[nDynamicObject] = 0;
		}

		m_nMode = 0;
		m_bHelp = true;
		m_bInvertMouse = true;
		Reset();
		
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Finalise()
	{
		if (m_ppDynamicObjectList)
		{
			for (int i = 0; i < m_nMaxNumDynamicObjects; i++)
			{
				if (m_ppDynamicObjectList[i])
				{
					m_ppDynamicObjectList[i]->Release();
					m_ppDynamicObjectList[i] = 0;
				}
			}
			delete []m_ppDynamicObjectList;
			m_ppDynamicObjectList = 0;
			m_nMaxNumDynamicObjects = 0;
			m_nNumDynamicObjects = 0;
		}
		
		if (m_ppCollisionObjectCombo)
		{
			for (int i = 0; i < NUM_COLLISION_OBJECTS; i++)
			{
				if (m_ppCollisionObjectCombo[i])
				{
					m_ppCollisionObjectCombo[i]->Release();
					m_ppCollisionObjectCombo[i] = 0;
				}
			}
			delete []m_ppCollisionObjectCombo;
			m_ppCollisionObjectCombo = 0;
		}
		

		if (m_pLandscape)
		{
			delete m_pLandscape;
			m_pLandscape = 0;
		}
		if (m_pMouse)
		{
			m_pMouse->Release();
			m_pMouse = 0;
		}
		if (m_pCamera)
		{
			m_pCamera->Release();
			m_pCamera = 0;
		}
		m_pApp = 0;

		Physics::DestroyInstance();

		DemoPhysicsRender::Finalise();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Update(float fDt)
	{
		//fDt = 1.0f / 60.0f;
		// Update Input
		m_pMouse->Update();

		if (m_pMouse->RightButtonJustPressed() || KeyIsJustPressed('K'))
		{
			m_nMode--;
			if (m_nMode < 0)
				m_nMode = 28;
			Reset();
		}
		else if (m_pMouse->LeftButtonJustPressed() || KeyIsJustPressed('L') )
		{
			m_nMode++;
			if (m_nMode > 28)
				m_nMode = 0;
			Reset();
		}
		else if (KeyIsJustPressed(VK_SPACE))
		{
			Reset();
		}


		if (m_nMode == 0 && m_nNumDynamicObjects < m_nMaxNumDynamicObjects)
		{	
			m_fTimer += fDt;
			float fTime = Exp(-m_nNumDynamicObjects * 0.01f) + 0.03f;
			while (m_fTimer > fTime)
			{
				m_fTimer -= fTime;
				m_pCamera->SetViewDistance(10.0f + m_nNumDynamicObjects * 0.01f);
				Vec3 v3Pos;
				v3Pos.x = (float)Cos(m_nNumDynamicObjects * 1.307f) * (1.5f + Sqrt(m_nNumDynamicObjects * 0.004f));
				v3Pos.y = 5.0f + Sqrt((float)m_nNumDynamicObjects * 0.07f);
				v3Pos.z = (float)Sin(m_nNumDynamicObjects * 1.113f) * (1.5f + Sqrt(m_nNumDynamicObjects * 0.007f));
				DynamicObject* pObject = AddObject(COLLISION_BOX, v3Pos);
				if (pObject)
				{
					pObject->SetLinearVelocity(Vec3(0.0f, -10.0f, 0.0f));
					pObject->SetAngularVelocity(Vec3(
						RandFunc::RandBetweenPosNeg(2.5f),
						RandFunc::RandBetweenPosNeg(2.5f),
						RandFunc::RandBetweenPosNeg(2.5f)));
				}
				if (m_nNumDynamicObjects >= m_nMaxNumDynamicObjects)
					break;
			    fTime = Exp(-m_nNumDynamicObjects * 0.01f) + 0.03f;
			}
		}

		if (KeyIsJustPressed(VK_F1))
			m_bHelp = !m_bHelp;
		if (KeyIsJustPressed('I'))
			m_bInvertMouse = !m_bInvertMouse;
		
		// Update physics
		Physics::GetInstance().Update(fDt);

		// Update the camera.	
		m_pCamera->SetNearClip(0.5f);
		m_pCamera->SetFarClip(1000.0f);
		m_pCamera->UpdateRotation(
			m_pMouse->GetXMovement(),
			m_bInvertMouse ? m_pMouse->GetYMovement() : -m_pMouse->GetYMovement());
		m_pCamera->SetTargetPosition(m_v3CameraTarget);
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Render()
	{
		// Set the camera
		m_pCamera->Realise();
		
		// Render physics (renders collision objects)
		Physics::GetInstance().Render();

		// Render the Landscape
		m_pLandscape->Render();

		RenderHelp();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void RenderHelp()
	{
		if (m_bHelp)
		{
			m_pApp->RenderText("Toggle help on and off (F1)");
			m_pApp->RenderText("Move the mouse to rotate the camera");
			m_pApp->RenderText("Invert mouse (I)");
			m_pApp->RenderText("Reset scene (Space)");
			m_pApp->RenderText("Next scene (Left Mouse Button) or (L)");
			m_pApp->RenderText("Previous scene (Right Mouse Button) or (K)");
			m_pApp->RenderText("Quit (Esc)");
			m_pApp->RenderText("");
			if (m_nMode == 0)
			{
				m_pApp->RenderText("Wait a while and the objects");
				m_pApp->RenderText("will start dropping faster.");
			}
		}
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	DynamicObject* AddObject(ObjectType eType, const MFrame& frame, const Vec3& v3Velocity = Vec3(0.0f, 0.0f, 0.0f))
	{
		if (m_nNumDynamicObjects < m_nMaxNumDynamicObjects)
		{
			DynamicObject* pDynamicObject = DynamicObject::CreateNew();
			pDynamicObject->Initialise(m_ppCollisionObjectCombo[eType]);
			pDynamicObject->SetFrame(frame);
			Physics::GetInstance().AddDynamicObject(pDynamicObject);	
			pDynamicObject->SetLinearVelocity(v3Velocity);
			pDynamicObject->SetMass(10.0f);
			m_ppDynamicObjectList[m_nNumDynamicObjects++] = pDynamicObject;
			return pDynamicObject;
		}
		return 0;
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Reset()
	{	
		m_fTimer = 0.0f;
		Vec3 v3GroundPos(1.0f, 0.0f, 1.0f);
		// Get placement height
		Collision collision = Physics::GetInstance().TestLineForCollision(
			v3GroundPos + k_v3UnitY * 1000.0f,
			-k_v3UnitY * 2000.0f,
			Physics::FLAG_ALL_NON_DYNAMIC_OBJECTS);

		if (collision.CollisionOccurred())
			v3GroundPos = collision.GetPosition();
		v3GroundPos.y += 1.0f;
		
		for (int nDynamicObject = 0; nDynamicObject < m_nNumDynamicObjects; nDynamicObject++)
		{
			Physics::GetInstance().RemoveDynamicObject(m_ppDynamicObjectList[nDynamicObject]);
			m_ppDynamicObjectList[nDynamicObject]->Release();
			m_ppDynamicObjectList[nDynamicObject] = 0;
		}
		m_nNumDynamicObjects = 0;

		Physics::GetInstance().SetMaxNumCollisions(4096);
		Physics::GetInstance().SetMaxNumMovingObjects(1024);

		// Initialise objects
		switch (m_nMode)
		{
		case 0:
			{	
				// Reduce the max number of collisions for mode 0 for a bit of extra performance
				// as we drop 1000s of objects onto the pile in quick sucession.
				Physics::GetInstance().SetMaxNumCollisions(1024);
				Physics::GetInstance().SetMaxNumMovingObjects(256);

				m_pCamera->SetViewDistance(10.0f);
				for (int nDynamicObject = 0; nDynamicObject < 16; nDynamicObject++)
				{
					Vec3 v3Pos = v3GroundPos;
					v3Pos.y += 1.98f * (nDynamicObject / 4);
					v3Pos.x += 0.75f * (2 * (nDynamicObject % 4) + ((nDynamicObject / 4) % 2));
					v3Pos.x += (float)RandFunc::RandBetweenPosNeg(0.1f);
					v3Pos.z += (float)RandFunc::RandBetweenPosNeg(0.1f);
					AddObject(COLLISION_BOX, v3Pos);
				}
			}
			break;

		case 1:
		case 2:
		case 3:
			{
				int nNumRows = 5;
				m_pCamera->SetViewDistance(10.0f);
				if (m_nMode == 2)
				{
					m_pCamera->SetViewDistance(20.0f);
					nNumRows = 10;
				}
				else
				if (m_nMode == 3)
				{
					m_pCamera->SetViewDistance(40.0f);
					nNumRows = 20;
				}
				for (int nRow = nNumRows; nRow >= 1; nRow--)
				{
					for (int nCol = 0; nCol < nRow; nCol++)
					{
						Vec3 v3Pos = v3GroundPos;
						v3Pos.y += 1.5f * (float)(nNumRows - nRow);
						v3Pos.x += 0.78f * (float)nCol * 2.0f + 0.78f * (float)(nNumRows - nRow - nNumRows * 0.75f);
						AddObject(COLLISION_BRICK, v3Pos);
					}
				}	
			}
			break;

		case 4:
		case 5:
		case 6:
			{
				int nNumRows = 5;
				m_pCamera->SetViewDistance(10.0f);
				if (m_nMode == 5)
				{
					m_pCamera->SetViewDistance(20.0f);
					nNumRows = 10;
				}
				else
				if (m_nMode == 6)
				{
					m_pCamera->SetViewDistance(40.0f);
					nNumRows = 20;
				}
				for (int nStacks = 0; nStacks < 2; nStacks++)
				{
					for (int nRow = nNumRows; nRow >= 1; nRow--)
					{
						for (int nCol = 0; nCol < nRow; nCol++)
						{
							Vec3 v3Pos = v3GroundPos;
							v3Pos.y += 1.5f * (float)(nNumRows - nRow);
							v3Pos.x += 0.78f * (float)nCol * 2.0f + 0.78f * (float)(nNumRows - nRow - nNumRows * 0.75f);
							v3Pos.z += (float)nStacks * 5.0f;
							AddObject(COLLISION_BRICK, v3Pos);
						}
					}	
				}
			}
			break;

		case 7:
		case 8:
		case 9:
		case 10:
		case 11:
			{
				int nNumRows = 5;
				Vec3 v3Offset(0.0f, 0.0f, 0.0f);
				m_pCamera->SetViewDistance(15.0f);
				if (m_nMode == 8)
				{
					m_pCamera->SetViewDistance(20.0f);
					nNumRows = 10;
				}
				else
				if (m_nMode == 9)
				{
					m_pCamera->SetViewDistance(40.0f);
					nNumRows = 20;
					v3Offset.x += RandFunc::RandBetweenPosNeg(10.0f);
				}
				else
				if (m_nMode == 10)
				{
					m_pCamera->SetViewDistance(40.0f);
					nNumRows = 30;
					v3Offset.x += RandFunc::RandBetweenPosNeg(10.0f);
				}
				else
				if (m_nMode == 11)
				{
					m_pCamera->SetViewDistance(50.0f);
					nNumRows = 60;
					v3Offset.x += RandFunc::RandBetweenPosNeg(10.0f);
				}
				for (int nRow = nNumRows; nRow >= 1; nRow--)
				{
					for (int nCol = 0; nCol < nRow; nCol++)
					{
						Vec3 v3Pos = v3GroundPos;
						// 0.78f is chosen here so that the objects stack perfectly as the would
						// if the where let to come to rest naturally.
						v3Pos.y += 0.78f * (float)(nNumRows - nRow - 1);
						v3Pos.x += 0.74f * (float)nCol * 2.0f + 0.74f * (float)(nNumRows - nRow - nNumRows * 0.74f);
						DynamicObject* pObject = AddObject(COLLISION_BRICK, v3Pos);
						if (pObject)
							pObject->SetToResting();
					}
				}				
				DynamicObject* pDynamicObject = 
					AddObject(COLLISION_BIG_BALL, v3GroundPos + Vec3(0.0f, 4.0f, 40.0f)+ v3Offset, Vec3(0.0f, 0.0f, -100.0f));
				if (pDynamicObject)
					pDynamicObject->SetMass(1000.0f);
			}
			break;
				
		case 12:
		case 13:
		case 14:
			{
				m_pCamera->SetViewDistance(10.0f);
				float fHeight = 0.0f;
				int nRows = 3;
				if (m_nMode == 13)
				{
					nRows = 5;
					m_pCamera->SetViewDistance(20.0f);
				}
				else
				if (m_nMode == 14)
				{
					nRows = 7;
					m_pCamera->SetViewDistance(20.0f);
				}
				for (int i = nRows; i > 0; i--)
				{
					for (int nDynamicObject = 0; nDynamicObject < i * i; nDynamicObject++)
					{						
						Vec3 v3Pos = v3GroundPos;
						v3Pos.x += 1.25f * (float)(nDynamicObject % i) - 1.25f * 0.5f * (float)i;
						v3Pos.y += fHeight;
						v3Pos.z += 1.25f * (float)(nDynamicObject / i) - 1.25f * 0.5f * (float)i;
						AddObject(COLLISION_BOX, v3Pos);
					}
					fHeight += 1.98f;
				}
			}
			break;

		case 15:
			{
				m_pCamera->SetViewDistance(20.0f);
				for (int nDynamicObject = 0; nDynamicObject < 32; nDynamicObject++)
				{
					Vec3 v3Pos = v3GroundPos;
					const int k_nStackHeight = 8;
					v3Pos.y += 
						(((nDynamicObject % (k_nStackHeight * 4)) * 0.1f + 1.98f) * 
						(nDynamicObject % k_nStackHeight));
					v3Pos.x += 4.0f * (((nDynamicObject / k_nStackHeight) % 4) - 1.5f);
					v3Pos.z += 4.0f * ((nDynamicObject / (k_nStackHeight * 4)));

					v3Pos.x += (float)RandFunc::RandBetweenPosNeg(0.1f);
					v3Pos.z += (float)RandFunc::RandBetweenPosNeg(0.1f);
					AddObject(COLLISION_BOX, v3Pos);
				}
			}
			break;

		case 16:
		case 17:
			{
				m_pCamera->SetViewDistance(15.0f);
				for (int nDynamicObject = 0; nDynamicObject < 64; nDynamicObject++)
				{
					Vec3 v3Pos = v3GroundPos;
					const int k_nStackHeight = 8;
					v3Pos.y += 0.98f * (nDynamicObject % k_nStackHeight);
					v3Pos.x += 4.0f * (((nDynamicObject / k_nStackHeight) %4) - 1.5f);
					v3Pos.z += 4.0f * ((nDynamicObject / (k_nStackHeight * 4)) - 0.5f);
					if (m_nMode == 17)
					{
						v3Pos.x += (float)RandFunc::RandBetweenPosNeg(0.1f);
						v3Pos.z += (float)RandFunc::RandBetweenPosNeg(0.1f);
					}
					DynamicObject* pDynamicObject = AddObject(COLLISION_BOX, v3Pos);
					if (m_nMode == 12 && pDynamicObject)
					{
						pDynamicObject->SetLinearVelocity(
							Vec3(RandFunc::RandBetweenPosNeg(1.5f), RandFunc::RandBetweenPosNeg(1.5f), RandFunc::RandBetweenPosNeg(0.15f)));
					}
				}
			}
			break;
		
		case 18:
		case 19:
		case 20:
			{
				m_pCamera->SetViewDistance(20.0f);
				int nCount = 8;
				if (m_nMode == 19)
				{				
					nCount = 24;	
					m_pCamera->SetViewDistance(30.0f);
				}
				else
				if (m_nMode == 20)
				{					
					m_pCamera->SetViewDistance(50.0f);
					nCount = 100;
				}
				for (int nDynamicObject = 0; nDynamicObject < nCount; nDynamicObject++)
				{
					Vec3 v3Pos = v3GroundPos;
					int nColumns = 4;
					Mat33 rotate(Mat33::ROTATION_AXIS_Y, k_fPi * 0.5f * nDynamicObject);
					v3Pos.y += 3 * (nDynamicObject / nColumns) + 2.0f;
					v3Pos.z += 1.4f;
					v3Pos.x -= 0.6f;
					AddObject(COLLISION_LONG_BOX, MFrame(v3Pos, Mat33(Mat33::ROTATION_AXIS_Z, k_fPi * 0.035f)) * rotate);
				}

			}
			break;
		
		case 21:
		case 22:
		case 23:
			{
				m_pCamera->SetViewDistance(20.0f);
				int nCount = 8;
				if (m_nMode == 22)
				{				
					nCount = 24;	
					m_pCamera->SetViewDistance(30.0f);
				}
				else
				if (m_nMode == 23)
				{					
					m_pCamera->SetViewDistance(50.0f);
					nCount = 100;
				}
				for (int nDynamicObject = 0; nDynamicObject < nCount; nDynamicObject++)
				{
					Vec3 v3Pos = v3GroundPos;
					int nColumns = 4;
					Mat33 rotate(Mat33::ROTATION_AXIS_Y, k_fPi * 0.5f * nDynamicObject);
					v3Pos.y += 1.0f * (nDynamicObject / nColumns) + 2.0f;
					v3Pos.z += 1.4f;
					v3Pos.x -= 0.6f;
					AddObject(COLLISION_LONG_CAPSULE, MFrame(v3Pos, Mat33(Mat33::ROTATION_AXIS_Z, k_fPi * 0.035f)) * rotate);
				}

			}
			break;

		case 24:
		case 25:
			{
				m_pCamera->SetViewDistance(15.0f);
				for (int nDynamicObject = 0; nDynamicObject < 64; nDynamicObject++)
				{
					Vec3 v3Pos = v3GroundPos;
					const int k_nStackHeight = 8;
					v3Pos.y += 0.98f * (nDynamicObject % k_nStackHeight);
					v3Pos.x += 4.0f * (((nDynamicObject / k_nStackHeight) %4) - 1.5f);
					v3Pos.z += 4.0f * ((nDynamicObject / (k_nStackHeight * 4)) - 0.5f);
					if (m_nMode == 25)
					{
						v3Pos.x += (float)RandFunc::RandBetweenPosNeg(0.1f);
						v3Pos.z += (float)RandFunc::RandBetweenPosNeg(0.1f);
					}
					DynamicObject* pDynamicObject = AddObject(COLLISION_CYLINDER, v3Pos);
					if (m_nMode == 20 && pDynamicObject)
					{
						pDynamicObject->SetLinearVelocity(
							Vec3(RandFunc::RandBetweenPosNeg(1.5f), RandFunc::RandBetweenPosNeg(1.5f), RandFunc::RandBetweenPosNeg(0.15f)));
					}
				}
			}
			break;

		case 26:
			{
				m_pCamera->SetViewDistance(15.0f);
				Vec3 v3Pos = v3GroundPos;
				AddObject(COLLISION_BOX, v3Pos);
				v3Pos.y += 1.2f;
				AddObject(COLLISION_BOX, v3Pos);
				v3Pos.y += 1.2f;
				AddObject(COLLISION_LONG_BOX, v3Pos);
				v3Pos.x += 2.0f;
				v3Pos.y += 2.0f;
				AddObject(COLLISION_BOX, v3Pos);
				v3Pos.x -= 4.0f;
				v3Pos.y += 5.0f;
				AddObject(COLLISION_BOX, v3Pos);
			}
			break;

		case 27:
			{
				m_pCamera->SetViewDistance(15.0f);
				Vec3 v3Pos = v3GroundPos;
				AddObject(COLLISION_BOX, v3Pos);
				v3Pos.y += 1.2f;
				DynamicObject* pDynamicObject = AddObject(COLLISION_LONG_BOX, v3Pos);
				pDynamicObject->SetFriction(3.0f);
				v3Pos.x += 2.0f;
				v3Pos.y += 1.2f;
				pDynamicObject = AddObject(COLLISION_BOX, v3Pos);
				pDynamicObject->SetFriction(3.0f);
				v3Pos.x -= 4.0f;
				v3Pos.y += 40.0f;
				pDynamicObject = AddObject(COLLISION_BOX, v3Pos);
				pDynamicObject->SetFriction(3.0f);
				pDynamicObject->SetMass(pDynamicObject->GetMass() * 2.0f);
			}
			break;

		case 28:
			{
				m_pCamera->SetViewDistance(20.0f);
				Vec3 v3Pos = v3GroundPos;
				for (int i = 0; i < 80; i++)
				{
					float fPos = (float)i * 0.52f;
					float fRadius = Cos((float)i * 0.94f) * 6.0f;

					Vec3 v3Offset(
						Sin(fPos) * fRadius, 
						0.0f, 
						Cos(fPos) * fRadius);

					MFrame frame;
					frame.v3Translation = v3Pos + v3Offset;
					frame.m33Rotation.Initialise(EulerAngles(
						RandFunc::RandUnit() * k_fPi, 
						RandFunc::RandUnit() * k_fPi, 
						RandFunc::RandUnit() * k_fPi));

					switch (RandFunc::RandBetween(0, 4))
					{
					case 0:
						AddObject(COLLISION_BOX, frame);
						break;

					case 1:
						AddObject(COLLISION_CYLINDER, frame);
						break;

					case 2:
						AddObject(COLLISION_LONG_BOX2, frame);
						break;

					case 3:
						AddObject(COLLISION_CAPSULE, frame);
						break;
					}

					v3Pos.y += 0.2f;
				}
			}
			break;
		}
	}

private:	
	IApp* m_pApp;
	IMouse* m_pMouse;	
	ICamera* m_pCamera;
	Landscape* m_pLandscape;
	CollisionObjectCombo** m_ppCollisionObjectCombo;

	int m_nMaxNumDynamicObjects;
	int m_nNumDynamicObjects;
	DynamicObject** m_ppDynamicObjectList;
	float m_fTimer;
	Vec3 m_v3CameraTarget;
	int m_nMode;
	bool m_bHelp;
	bool m_bInvertMouse;
};

TA_IMPLEMENT_DEMO_PLUGIN(DemoStackingObjects);
