//---------------------------------------------------------------------------------
// File Name: HeightField.cpp
// Description:
//
// Copyright (C) 2002 True Axis Pty Ltd, Australia. 
// All Rights Reserved.
//
// Use of this code is covered by the "TRUE AXIS PHYSICS SDK DEMOS SOURCE LICENSE AGREEMENT".
// Details of this agreement are found in the file "TrueAxisPhysicsSDKDemosSourceLicence.txt".
// You have permission to modify this file and use it for any purpose under the terms of
// the agreement.
//
// History:
//		Created File.
//---------------------------------------------------------------------------------

#include "HeightField.h"

#ifndef TA_MATHS_H
#include "../TA/Common/Maths.h"
#endif // TA_MATHS_H

#ifndef TA_RANDFUNC_H
#include "../TA/Common/RandFunc.h"
#endif // TA_RANDFUNC_H

#include <memory.h>

using namespace TA;

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
HeightField::HeightField()
{	
	m_pfHeightData = 0;
	m_nSizeX = 0;
	m_nSizeY = 0;
	m_fHorizontalScale = 0.0f;
	m_fVerticalScale = 0.0f;
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
HeightField::~HeightField()
{
	Finalise();
}

//---------------------------------------------------------------------------------
// gausian random number between 1 and -1
//---------------------------------------------------------------------------------
inline float HeightField::Guassian()
{
	float fSum = 0.0f;
	for (int i = 0; i < 10; i++)
		fSum += RandFunc::RandUnit();
	return fSum * (1.0f / 5.0f) - 1.0f;
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void HeightField::Initialise(
	int nLevelsX, 
	int nLevelsY, 
	float fFractalDimension, 
	int nSeed, 
	float fHorizontalScale,
	float fVerticalScale,
	int nSmoothAmount,
	int nQuatize)
{
	m_fHorizontalScale = fHorizontalScale;
	m_fVerticalScale = fVerticalScale;

	TA_ASSERT(nLevelsX == nLevelsY);
    int nLevel = 0;
	
	if (m_pfHeightData)
	{
		TA_ASSERT(0);
		Finalise();
	}

	RandFunc::RandSeed(nSeed);

	int nSizeX = 1 << nLevelsX;
    int nSizeY = 1 << nLevelsY;
	m_nSizeX = nSizeX + 1;
	m_nSizeY = nSizeY + 1;
	
	m_pfHeightData = new float[m_nSizeX * m_nSizeY];
	memset(m_pfHeightData, 0, sizeof (float) * m_nSizeX * m_nSizeY);

    if (nSizeX == nSizeY)
    {
		HeightFeild(0, 0) =						
			Guassian() / Pow(fFractalDimension, (float)nLevel);
		HeightFeild(0, nSizeX) =				
			Guassian() / Pow(fFractalDimension, (float)nLevel);
		HeightFeild(0, nSizeX >> 1) =			
			Guassian() / Pow(fFractalDimension, (float)nLevel + 1.0f);

		HeightFeild(nSizeY, 0) =					
			Guassian() / Pow(fFractalDimension, (float)nLevel);
		HeightFeild(nSizeY, nSizeX) =			
			Guassian() / Pow(fFractalDimension, (float)nLevel);
		HeightFeild(nSizeY, nSizeX >> 1) =		
			Guassian() / Pow(fFractalDimension, (float)nLevel + 1.0f);

		HeightFeild(nSizeY >> 1, 0) =	
			Guassian() / Pow(fFractalDimension, (float)nLevel + 1.0f);
		HeightFeild(nSizeY >> 1, nSizeX) =		
			Guassian() / Pow(fFractalDimension, (float)nLevel + 1.0f);
		HeightFeild(nSizeY >> 1, nSizeX >> 1) =			
			Guassian() / Pow(fFractalDimension, (float)nLevel + 1.0f);

        DoGenerateLandscape(nLevelsX, 0, fFractalDimension);

		if (nQuatize)
		{
			for (int nX = 0; nX < m_nSizeX; nX++)
			{
				for (int nY = 0; nY < m_nSizeX; nY++)
				{
					m_pfHeightData[m_nSizeX * nY + nX] = 
						(float)((int)(m_pfHeightData[m_nSizeX * nY + nX] * (float)nQuatize)) / 
						(float)nQuatize;
				}
			}
		}

	
		if (nSmoothAmount)
		{
			int nSampleSize = nSmoothAmount;
			float* pfHeightDataNew = new float[m_nSizeX * m_nSizeY];
			memset(pfHeightDataNew, 0, sizeof (float) * m_nSizeX * m_nSizeY);
			for (int nX = 0; nX < m_nSizeX; nX++)
			{
				for (int nY = 0; nY < m_nSizeX; nY++)
				{
					int nCount = 0;
					float fSum = 0.0f;
					int nMinSX = nX - nSampleSize;
					if (nMinSX < 0)
						nMinSX = 0;
					int nMaxSX = nX + nSampleSize;
					if (nMaxSX > m_nSizeX - 1)
						nMaxSX = m_nSizeX - 1;
					int nMinSY = nY - nSampleSize;
					if (nMinSY < 0)
						nMinSY = 0;
					int nMaxSY = nY + nSampleSize;
					if (nMaxSY > m_nSizeY - 1)
						nMaxSY = m_nSizeY - 1;
					for (int nSX = nMinSX; nSX <= nMaxSX; nSX++)
					{
						for (int nSY = nMinSY; nSY <= nMaxSY; nSY++)
						{
							fSum += m_pfHeightData[m_nSizeX * nSY + nSX];
							nCount++;
						}
					}
					
					TA_ASSERT(nCount > 0);
					pfHeightDataNew[m_nSizeX * nY + nX] = fSum / (float)nCount;
				}
			}
			delete m_pfHeightData;
			m_pfHeightData = pfHeightDataNew;	
		}
	}

}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void HeightField::Initialise(
	float* pfData,
	int nWidth, 
	int nHeight,
	float fHorizontalScale,
	float fVerticalScale)
{
	m_fHorizontalScale = fHorizontalScale;
	m_fVerticalScale = fVerticalScale;
	
    int nLevel = 0;
	
	if (m_pfHeightData)
	{
		TA_ASSERT(0);
		Finalise();
	}

	m_nSizeX = nWidth;
	m_nSizeY = nHeight;
	
	m_pfHeightData = new float[m_nSizeX * m_nSizeY];
	for (int i = 0; i < m_nSizeY; i++)
	{
		for (int j = 0; j < m_nSizeX; j++)
		{
			HeightFeild(i, j) =	pfData[m_nSizeX * i  + j];
		}
	}
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void HeightField::Finalise()
{
	if (m_pfHeightData)
	{
		delete []m_pfHeightData;
		m_pfHeightData = 0;
	}
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void HeightField::DoGenerateLandscape(int nLevels, int nLevel, float fFractalDimension)
{
	int nNewPointX;
	int nNewPointY;
	int nX, nY;
	int nSize;
	int nInc;
	float nNewVal ;
	int nStart;
    int nX1;
	int nY1;
	int nX2;
	int nY2;

	nSize = 1 << nLevels;

	nStart = nSize >> 2;
	nInc = nSize >> 1;

	for (;;)
	{
		float GausianMult = 1.0f / Pow(fFractalDimension, (float)nLevel);
		//---------------------------------------------------------------------------------
		// pass 1
		//---------------------------------------------------------------------------------
		for (nY = nStart; nY < nSize; nY += nInc)
		{
			nY1 = nY - nStart; nY2 = nY + nStart;
			nNewPointY = nY;
			for (nX = nStart; nX <= nSize; nX += nInc)
			{
				nX1 = nX - nStart; nX2 = nX + nStart;

				nNewPointX = nX;
				nNewVal = 0;
				nNewVal += HeightFeild(nY1, nX1);
				TA_ASSERT(nNewVal > - 10.0f);
				nNewVal += HeightFeild(nY1, nX2);
				TA_ASSERT(nNewVal > - 10.0f);
				nNewVal += HeightFeild(nY2, nX1);
				TA_ASSERT(nNewVal > - 10.0f);
				nNewVal += HeightFeild(nY2, nX2);
				TA_ASSERT(nNewVal > - 10.0f);
				nNewVal /= 4.0f;
				
				nNewVal += Guassian() * GausianMult;
				HeightFeild(nNewPointY, nNewPointX) = nNewVal;
			}
		}
		//---------------------------------------------------------------------------------
		// pass 2
		//---------------------------------------------------------------------------------
		//---------------------------------------------------------------------------------
		// part1
		//---------------------------------------------------------------------------------
		for (nY=0; nY<nSize; nY+=nInc)
		{
			nY1 = nY - nStart; nY2 = nY + nStart;
			nNewPointY = nY;
			for (nX = nStart; nX <= nSize; nX += nInc)
			{
				nX1 = nX - nStart; nX2 = nX + nStart;
				nNewPointX=nX;
				nNewVal =0;
				if (nX == 0)
				{
					nNewVal += HeightFeild(nY1, nNewPointX);
					nNewVal += HeightFeild(nY2, nNewPointX);
					nNewVal += HeightFeild(nNewPointY, nX2);
					nNewVal /=3.0;
				}
				else if (nX == nSize)
				{
					nNewVal += HeightFeild(nY1, nNewPointX);
					nNewVal += HeightFeild(nY2, nNewPointX);
					nNewVal += HeightFeild(nNewPointY, nX1);
					nNewVal /=3.0;
				}
				else if (nY == 0)
				{
					nNewVal += HeightFeild(nY2, nNewPointX);
					nNewVal += HeightFeild(nNewPointY, nX1);
					nNewVal += HeightFeild(nNewPointY, nX2);
					nNewVal /= 3.0;
				}
				else if (nY==nSize)
				{
					nNewVal += HeightFeild(nY1, nNewPointX);
					nNewVal += HeightFeild(nNewPointY, nX1);
					nNewVal += HeightFeild(nNewPointY, nX2);
					nNewVal /= 3.0;
				}
				else
				{
					nNewVal += HeightFeild(nY1, nNewPointX);
					nNewVal += HeightFeild(nY2, nNewPointX);
					nNewVal += HeightFeild(nNewPointY, nX1);
					nNewVal += HeightFeild(nNewPointY, nX2);
					nNewVal /= 4.0;
				}
				TA_ASSERT(nNewVal > - 10.0f);
				
				nNewVal += Guassian() * GausianMult;
				HeightFeild(nNewPointY, nNewPointX) = nNewVal;
			}
		}
		//---------------------------------------------------------------------------------
		// part 2
		//---------------------------------------------------------------------------------
		for (nY=nStart; nY<nSize; nY+=nInc)
		{
			nY1=nY-nStart; nY2=nY+nStart;
			nNewPointY=nY;
			for (nX = 0; nX <= nSize; nX += nInc)
			{
				nX1 = nX - nStart; nX2 = nX + nStart;
				nNewPointX = nX;
				nNewVal = 0;
				if (nX == 0)
				{
					nNewVal += HeightFeild(nY1, nNewPointX);
					nNewVal += HeightFeild(nY2, nNewPointX);
					nNewVal += HeightFeild(nNewPointY, nX2);
					nNewVal /= 3.0;
				}
				else if (nX == nSize)
				{
					nNewVal += HeightFeild(nY1, nNewPointX);
					nNewVal += HeightFeild(nY2, nNewPointX);
					nNewVal += HeightFeild(nNewPointY, nX1);
					nNewVal /= 3.0;
				}
				else if (nY == 0)
				{
					nNewVal += HeightFeild(nY2, nNewPointX);
					nNewVal += HeightFeild(nNewPointY, nX1);
					nNewVal += HeightFeild(nNewPointY, nX2);
					nNewVal /= 3.0;
				}
				else if (nY == nSize)
				{
					nNewVal += HeightFeild(nY1, nNewPointX);
					nNewVal += HeightFeild(nNewPointY, nX1);
					nNewVal += HeightFeild(nNewPointY, nX2);
					nNewVal /= 3.0;
				}
				else
				{
					nNewVal += HeightFeild(nY1, nNewPointX);
					nNewVal += HeightFeild(nY2, nNewPointX);
					nNewVal += HeightFeild(nNewPointY, nX1);
					nNewVal += HeightFeild(nNewPointY, nX2);
					nNewVal /= 4.0;
				}
				TA_ASSERT(nNewVal > - 10.0f);
				nNewVal += Guassian() * GausianMult;
				HeightFeild(nNewPointY, nNewPointX) = nNewVal;
			}
		}

		nStart /= 2;
		nInc /= 2;
		if (nStart < 1) 
			break;
		nLevel++;
	}
}
